package xylia.sax.contentmodelparser;

import java.util.* ;
import xylia.utilities.Assert;

/** Converts a regular expression to a Deterministe Finite Automaton (DFA).
 *  The alogrithm used is that presented in Aho, Sethi, and Ullman's second
 *  edition of Compilers: Principles, Techniques, and Tools, in section 3.9.
 *  The automaton is not minimized.
 *  
 * <p> The algorithm first converts a regular expression r to r# where #
 * is a unique symbol. A <em>position</em> is a node in r# that represents
 * a symbol. In XML terms, the symbols are element names or the special
 * symbol #PCDATA. Then the following functions are computed for each
 * node of r#.</P>
 * <ul><li> firstpos(s) The set of all positions in s that might match the
 * first symbol of a string.</li>
 *     <li> lastpos(s) The set of all positions in s that might match the
 * last symbol of a string.</li>
 *     <li> nullable(s) Whether s can recognize the empty string.</li>
 * </ul>
 * <p>Next we compute the following function on the postions of r#</p>
 * <ul><li> followpos(p) The set of all positons q in r# such that p and q
 *          might match successive symbols in a string.</li>
 * </ul>
 * Now followpos provides the structure of an NFA. We do a subset construction
 * on followpos to compute the states and transitions of a DFA as follows
 * (symbols is the set of all symbols excluding #).
 * <pre>
 *     const start := firstpos(r#)
 *     states := {start}
 *     var unmarked := {start}
 *     while unmarked not= {} 
 *         const T :in unmarked
 *         unmarked := unmarked \ {T}
 *         for a in symbols 
 *             const U := Union( p in T . {q in followpos(p) | q matches a})
 *             if U not= {}
 *                 states := states union {U}
 *                 delta(T,a) := U
 *     final := {S in states | S contains a position matching #} 
 * </pre>
 * <p>Now the DFA is (states, symbols, delta, start, final).</p>
 * <p>However the above construction creates states that are sets.
 * For efficiency, this class creates DFAs with Integers as states.</p>
 * 
 * <p>Copyright: Copyright (c) 2002 Theodore S. Norvell</p>
 * <p>Company: MUN</p>
 * @author Theodore S. Norvell
 * @version 1.0
 */
 
public class ConvertToDFA {

    static private String symbolFor( RegExp r ) {
        switch( r.getKind() ) {
            case RegExp.NAME : {
                return r.getName() ; }
            case RegExp.PCDATA : {
                return "#PCDATA" ; }
            default: Assert.assert(false) ; return null ; } }
            
    static private void firstpass( RegExp r, Map firstpos, Map lastpos,
                                   Set nullable, Map followpos, Set symbols ) {
        switch( r.getKind() ) {
            case RegExp.NAME : case RegExp.PCDATA : {
                followpos.put( r, new HashSet() ) ;
                Set ps = new TreeSet() ; ps.add( r ) ;
                firstpos.put( r, ps ) ;
                lastpos.put( r, ps ) ;
                symbols.add( symbolFor( r ) ) ; }
            break ;
            case RegExp.ALT : {
                firstpass( r.getOperand(0), firstpos, lastpos, nullable, followpos, symbols ) ;
                firstpass( r.getOperand(1), firstpos, lastpos, nullable, followpos, symbols ) ;
                
                // nullable(r) := nullable( r0 ) or nullable( r1 )
                if( nullable.contains( r.getOperand(0) )
                ||  nullable.contains( r.getOperand(1)) ) {
                    nullable.add(r) ; }
                
                // firstpos( r ) := firstpos( r0 ) U firstpos( r1 )
                Set ufpos = new HashSet() ;
                ufpos.addAll( (Set) firstpos.get( r.getOperand(0) ) ) ;
                ufpos.addAll( (Set) firstpos.get( r.getOperand(1) ) ) ;
                firstpos.put( r, ufpos ) ;
                
                // lastpos( r ) := lastpos( r0 ) U lastpos( r1 )
                Set ulpos = new HashSet() ;
                ulpos.addAll( (Set) lastpos.get( r.getOperand(0) ) ) ;
                ulpos.addAll( (Set) lastpos.get( r.getOperand(1) ) ) ;
                lastpos.put( r, ulpos ) ; }
            break ;
            case RegExp.SEQ : {
                firstpass( r.getOperand(0), firstpos, lastpos, nullable, followpos, symbols ) ;
                firstpass( r.getOperand(1), firstpos, lastpos, nullable, followpos, symbols ) ;
                
                // nullable( r ) := nullable( r0 ) and nullable( r1 )
                
                if( nullable.contains( r.getOperand(0) )
                &&  nullable.contains( r.getOperand(1)) ) {
                    nullable.add(r) ; }
                
                // firstpos(r) := firstpos( r0)
                //              U if nullable( r0 ) then firstpos( r1 ) else {}
                Set fpos = new HashSet() ;
                fpos.addAll( (Set) firstpos.get( r.getOperand(0) ) ) ;
                if( nullable.contains( r.getOperand(0) ) ) {
                    fpos.addAll( (Set) firstpos.get( r.getOperand(1) ) ) ; }
                firstpos.put( r, fpos ) ;
                
                // lastpos(r) := lastpos( r1)
                //             U if nullable( r1 ) then lastpos( r0 ) else {}
                Set lpos = new HashSet() ;
                lpos.addAll( (Set) lastpos.get( r.getOperand(1) ) ) ;
                if( nullable.contains( r.getOperand(1) ) ) {
                    lpos.addAll( (Set) lastpos.get( r.getOperand(0) ) ) ; }
                lastpos.put( r, lpos ) ; }
            break ;
            case RegExp.KSTAR : {
                firstpass( r.getOperand(0), firstpos, lastpos, nullable, followpos, symbols ) ;
                
                // nullable(r) := true
                nullable.add( r ) ;
                // firstpos(r) = firstpos(r0)
                firstpos.put( r, firstpos.get( r.getOperand(0) ) ) ;
                // lastpos(r) := lastpos(r0)
                lastpos.put( r, lastpos.get( r.getOperand(0) ) ) ; }
            break ;
            case RegExp.KPLUS : {
                firstpass( r.getOperand(0), firstpos, lastpos, nullable, followpos, symbols ) ;
                
                // nullable(r) := nullable(r0)
                if( nullable.contains( r.getOperand(0) ) ) {
                    nullable.add( r ) ; }
                
                // firstpos(r) := firstpos(r0)
                firstpos.put( r, firstpos.get( r.getOperand(0) ) ) ;
                // lastpos(r) := lastpos(r0)
                lastpos.put( r, lastpos.get( r.getOperand(0) ) ) ; }
            break ;
            case RegExp.OPT : {
                firstpass( r.getOperand(0), firstpos, lastpos, nullable, followpos, symbols ) ;
                
                // nullable(r) := true
                nullable.add( r ) ;
                // firstpos(r) = firstpos(r0)
                firstpos.put( r, firstpos.get( r.getOperand(0) ) ) ;
                // lastpos(r) := lastpos(r0)
                lastpos.put( r, lastpos.get( r.getOperand(0) ) ) ; }
        }
    }
                
    static private void secondpass( RegExp r, Map firstpos, Map lastpos, Set nullable,
                                    Map followpos ) {
        switch( r.getKind() ) {
            case RegExp.NAME : case RegExp.PCDATA : {
                 }
            break ;
            case RegExp.ALT : {
                secondpass( r.getOperand(0), firstpos, lastpos, nullable, followpos ) ;
                secondpass( r.getOperand(1), firstpos, lastpos, nullable, followpos ) ;
                }
            break ;
            case RegExp.SEQ : {
                secondpass( r.getOperand(0), firstpos, lastpos, nullable, followpos ) ;
                secondpass( r.getOperand(1), firstpos, lastpos, nullable, followpos ) ;
                
                // followpos(p) := followpos(pos) U firstpos(r1), for each p in lastpos(r0)
                Set fpos_1 = (Set) firstpos.get( r.getOperand(1) ) ;
                Iterator i = ((Set)lastpos.get( r.getOperand(0) )).iterator() ;
                while( i.hasNext() ) {
                    RegExp p = (RegExp) i.next() ;
                    ((Set)followpos.get( p )).addAll( fpos_1 ) ; }
                 }
            break ;
            case RegExp.KSTAR : case RegExp.KPLUS :  {
                secondpass( r.getOperand(0), firstpos, lastpos, nullable, followpos ) ;
                
                // followpos(p) := followpos(p) U firstpos(r0), for each p in lastpos(r0)
                Set fpos_0 = (Set) firstpos.get( r.getOperand(0) ) ;
                Iterator i = ((Set)lastpos.get( r.getOperand(0) )).iterator() ;
                while( i.hasNext() ) {
                    RegExp p = (RegExp) i.next() ;
                    ((Set)followpos.get( p )).addAll( fpos_0 ) ; }
                }
            break ;
            case RegExp.OPT : {
                secondpass( r.getOperand(0), firstpos, lastpos, nullable, followpos ) ;
                }
        }
    }
                
                
                
                
    /** Implements the algorithm in of Aho Sethi and Ullman section 3.9
     *  converting a regular expression (unaugmented) to a dfa (not minimized).
     *  @param r a regular expression. r must be a tree (no shared nodes).
     */
    static public FiniteStateAutomaton convertToDFA( RegExp r ) {
        // Augment the regular Expression using "" as a stop symbol.
        RegExp finalPosn = new RegExp( RegExp.NAME, "" ) ;
        r = new RegExp( RegExp.SEQ, r, finalPosn ) ;
        
        //Compute nullable, firstpos, lastpos
        Map firstpos = new HashMap() ;
        Map lastpos = new HashMap() ;
        Set nullable = new HashSet() ;
        Map followpos = new HashMap() ;
        Set symbols = new HashSet() ;
        
        // Compute firstpos, lastpos, nullable, and symbols.
        // Initialize followpos(p) := {} for all positions.
        firstpass( r, firstpos, lastpos, nullable, followpos, symbols ) ;
        symbols.remove("") ;
        
        //Compute followpos
        secondpass( r, firstpos, lastpos, nullable, followpos ) ;
        
        // Compute the automaton.
        
        StateTransitionStructure sts = new StateTransitionStructure() ; 
        // We use Integers for states rather than
        // sets and keep the correspondence in the following two maps
        int stateCounter = 0 ;
        Map stateFor = new HashMap() ;
        Map setFor = new TreeMap() ;
        
        // The start State
        
        Set startSet = (Set) firstpos.get( r ) ;
        Integer startState = new Integer( stateCounter++ ) ;
        stateFor.put( startSet, startState ) ;
        setFor.put( startState, startSet ) ;
        sts.addState( startState );
        
        // Unmarked states are not yet dealt with
        Set unmarked = new TreeSet( ) ;
        unmarked.add( startState ) ;
        while( ! unmarked.isEmpty() ) {
            // Remove an element t from unmarked.
            Iterator it = unmarked.iterator() ;
            Integer t = (Integer) it.next() ;
            unmarked.remove( t ) ;
            Set T = (Set) setFor.get(t) ;
            
            //For each input symbol a
            Iterator symbol_it = symbols.iterator() ;
            while( symbol_it.hasNext() ) { 
                Object a = symbol_it.next() ;
                // Let U be the set of all positions in followpos(p),
                // for some position p in T such that symbol a is the
                // the symbol for p
                Set U = new HashSet() ;
                Iterator t_it = T.iterator() ;
                while( t_it.hasNext() ) {
                    RegExp p = (RegExp) t_it.next() ;
                    if( symbolFor(p).equals(a) ) {
                        U.addAll( (Set) followpos.get( p ) ) ; } }
            
                if( ! U.isEmpty() ) {
                    // If there is no state for U, create one
                    // and add it to unmarked
                    Integer u ;
                    if( stateFor.containsKey(U)) {
                        u = (Integer) stateFor.get(U) ; }
                    else {
                        u = new Integer( stateCounter++ ) ;
                        setFor.put( u, U ) ;
                        stateFor.put( U, u ) ;
                        sts.addState( u ) ;
                        unmarked.add( u ) ; }
                    // There is a transition t --a--> u
                    sts.addTransition(t, a, u) ; } } }
            
        // Compute the final state
        Set F = new TreeSet() ;
        // For each state, if its set contains the final position add it.
        Iterator state_it = setFor.keySet().iterator() ;
        while( state_it.hasNext() ) {
            Integer state = (Integer) state_it.next() ;
            Set set = (Set) setFor.get( state ) ;
            if( set.contains( finalPosn ) ) {
                F.add( state ) ; } }
            
        return new FiniteStateAutomaton( sts, startState, F ) ;
    }
}