package xylia.sax.contentmodelparser;

import java.util.* ;

import xylia.utilities.Assert ;

/** A content model able to check validity of a sequence of children.
 * Title:        Xylia
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:      Memorial
 * @author Theodore Norvell, Dennis Peters and others.
 * @version 1.0
 */

public class DefaultContentModel implements ContentModelKinds {

    private int kind = -1;
    private RegExp re = null ;
    private FiniteStateAutomaton fsa = null ;
    
    /** Construct an ANY or EMPTY content model. */
    public DefaultContentModel( int kind) {
        Assert.assert( kind == ANY
                    || kind == EMPTY ) ;
        this.kind = kind ;
    }

    /** Construct a CHILDREN or MIXED content model */
    public DefaultContentModel( int kind, RegExp re ) {
        Assert.assert( kind == CHILDREN
                    || kind == MIXED ) ;
        this.kind = kind ;
        this.re = re ;
    }

    /** Returns one of EMPTY, ANY, CHILDREN, or MIXED. */
    public int getKind() { return kind ; }

    /** Return a regular expression for the content model. 
     *  <p>Precondition: kind must be CHILREN or MIXED.</p>*/
    public RegExp getRegExp() {
        Assert.assert( kind == CHILDREN
                    || kind == MIXED ) ;
        return re ; }
    
    /** Return a finite state automaton for the content model
     *  <p>Precondition: kind must be CHILREN or MIXED.</p> 
     *  <P>The alphabet of the automaton is the element
     *  names (Strings) and the String "#PCDATA"</p>
     *  <p>Note that if the content model is (#PCDATA)
     *  then it is first converted to (#PCDATA)*, so the 
     *  automaton will match any number of "#PCDATA".</p>
     */
    public FiniteStateAutomaton getFSA() {
        Assert.assert( kind == CHILDREN
                    || kind == MIXED ) ;
        if( fsa == null ) {
            // Treat #PCDATA as (#PCDATA)* so that 
            // repeated (or absent) #PCDATA blocks are
            // considered valid.
            RegExp re0 ;
            if( kind == MIXED && re.getKind() == RegExp.PCDATA ) {
                re0 = new RegExp( RegExp.KSTAR, re ) ; }
            else {
                re0 = re ; }
            fsa = ConvertToDFA.convertToDFA( re0 ) ; }
        return fsa ; }
    
    /** Determine whether a sequence of content elements is matched by the automaton.
     *  @param content. A List of Strings. The Strings can be element names or "#PCDATA".
     */
    public boolean match( List content ) {
        switch( kind ) {
            case EMPTY: return content.size()==0 ;
            case ANY: return true ;
            default: return getFSA().match( content ) ; }
    }
}